module data.list {
   constructor: '=>'/2.

   syntax: infix '=>'.

   /** length(List, N).
    * Succeeds if <code>List</code> is a list with <code>N</code> elements.
    */

   length(List, N) :-
      'kernel':list_length(List, N).


   /** member(X, List).
    * Succeeds if <code>List</code> is a (partial) list whose <code>X</code> is an element.
    */

   member(H, [H | _]).
   member(H, [_ | T]) :-
      member(H, T).

   member_last(H, [_ | T]) :-
      member_last(H, T).
   member_last(H, [H | _]).

   add_to_left(Item, List) :-
      List = [Head | Tail],
      data.term:(
         setarg1(1, List, Item),
         setarg1(2, List, [Head | Tail])
      ).

   /** append(A, B, C).
    * Succeeds if <code>A</code> is a list and <code>C</code> is the
    * concatenation of <code>A</code> and <code>B</code>.
    */

   append([], L, L).
   append([H | T0], L, [H | T1]) :-
      append(T0, L, T1).


   syntax: infix right associative '++'.

   syntax: prefix '|' < '++' < ':'. 

   `++'(A, B) = append(A, B).

   syntax: postfix '*'.

   syntax: '++' < postfix '*' < ':'.

   `*'(X, Y) :-
      Y match ([]; X ++ X*).

   /** flatten(ListOfLists, List).
    * <code>flatten([L<sub><math>1</math></sub>, ...,
    *    L<sub><math>n</math></sub>], L)</code> succeeds if
    * <code>L</code> is the concatenation of the lists
    * <code>L<sub><math>1</math></sub></code>, ...,
    * <code>L<sub><math>n</math></sub></code>.
    */

   flatten([], []).
   flatten([H | T], L) :-
      append(H, LT, L),
      flatten(T, LT).

   /** join(ListOfLists, Delimiter, List).
    * <prerequisite>
    * <code>ListOfLists</code> and <code>Delimiter</code> are lists.
    * </prerequisite>
    * <code>join([L<sub><math>1</math></sub>, ...,
    *    L<sub><math>n</math></sub>], Delimiter, L)</code> succeeds if
    * <code>L</code> is the concatenation of
    * <code>L<sub><math>1</math></sub></code>, Delimiter,
    * <code>L<sub><math>2</math></sub></code>, Delimiter, ...,
    * <code>L<sub><math>n-1</math></sub></code>, Delimiter,
    * <code>L<sub><math>n</math></sub></code>.
    */

   join([], _, []).
   join([H | T], Delim, L) :-
      append(H, LT, L),
      join_with_prefix(T, Delim, LT).

   /** join_with_prefix(ListOfLists, Delimiter, List).
    * <prerequisite>
    * <code>ListOfLists</code> and <code>Delimiter</code> are lists.
    * </prerequisite>
    * <code>join_with_prefix([L<sub><math>1</math></sub>, ...,
    *    L<sub><math>n</math></sub>], Delimiter, L)</code> succeeds if
    * <code>L</code> is the concatenation of
    * Delimiter, <code>L<sub><math>1</math></sub></code>,
    * Delimiter, <code>L<sub><math>2</math></sub></code>, ...,
    * Delimiter, <code>L<sub><math>n</math></sub></code>.
    */

   join_with_prefix([], _, []).
   join_with_prefix([H | T], Delim, L) :-
      append(Delim, L0, L),
      append(H, L1, L0),
      join_with_prefix(T, Delim, L1).

   /** reverse(L0, L1).
    * Succeeds if the list <code>L0</code> unifies with
    * the list <code>L1</code> in reverse order.
    */

   reverse(L0, L1) :-
      reverse_append(L0, [], L1).

   /** reverse_append(A, B, C).
    * Succeeds if the list <code>C</code> is the
    * concatenation of the reverse of <code>A</code> and
    * <code>B</code>.
    */

   reverse_append([], L, L).
   reverse_append([H | T], L, R) :-
      reverse_append(T, [H | L], R).

   /** iter(List, F).
    * <prerequisite>
    * <code>F</code> is a module which implements a predicate <code>do(X)</code>.
    * </prerequisite>
    *  <code>iter([X<sub><math>1</math></sub>, ...,
    *    X<sub><math>n</math></sub>], F)</code> calls
    *  <code>F:do(X<sub><math>i</math></sub>)</code>
    *  for <math>i</math> from <math>1</math> to
    *  <math>n</math>.
    */

   iter([], _).
   iter([H | T], F) :-
      F:do(H),
      iter(T, F).

   iteri(L, F) :-
      iteri_from(L, 1, F).

   iteri_from([], _, _).
   iteri_from([H | T], N, F) :-
      F:do(H, N),
      iteri_from(T, data.number:succ(N), F).

   /** iter(List0, F, List1).
    * <prerequisite>
    * <code>F</code> is a module which implements a predicate <code>do(X, Y)</code>.
    * </prerequisite>
    *  <code>iter([X<sub><math>1</math></sub>, ...,
    *    X<sub><math>n</math></sub>], F,
    *   [Y<sub><math>1</math></sub>, ...,
    *    Y<sub><math>n</math></sub>])</code> calls
    *  <code>F:do(X<sub><math>i</math></sub>,
    *             Y<sub><math>i</math></sub>)</code>
    *  for <math>i</math> from <math>1</math> to
    *  <math>n</math>.
    */

   iter([], _, []).
   iter([H0 | T0], F, [H1 | T1]) :-
      F:do(H0, H1),
      iter(T0, F, T1).

   iteri(L, F) = iteri_from(L, 1, F).

   iteri_from([], _, _, []).
   iteri_from([H0 | T0], N, F, [H1 | T1]) :-
      F:do(H0, N, H1),
      iteri_from(T0, data.number:succ(N), F, T1).

   map(L, F) = iter(L, F).

   mapi(L, F) = iteri(L, F).

   /** iter(List0, List1, List2, F).
    * <prerequisite>
    * <code>F</code> is a module which implements a predicate <code>do(X, Y, Z)</code>.
    * </prerequisite>
    *  <code>iter([X<sub><math>1</math></sub>, ...,
    *    X<sub><math>n</math></sub>],
    *   [Y<sub><math>1</math></sub>, ...,
    *    Y<sub><math>n</math></sub>],
    *   [Z<sub><math>1</math></sub>, ...,
    *    Z<sub><math>n</math></sub>], F)</code> calls
    *  <code>F:do(X<sub><math>i</math></sub>,
    *             Y<sub><math>i</math></sub>,
    *             Z<sub><math>i</math></sub>)</code>
    *  for <math>i</math> from <math>1</math> to
    *  <math>n</math>.
    */

   iter([], [], [], _).
   iter([H0 | T0], [H1 | T1], [H2 | T2], F) :-
      F:do(H0, H1, H2),
      iter(T0, T1, T2, F).

   /** reverse_iter(List, F).
    * <prerequisite>
    * <code>F</code> is a module which implements a predicate <code>do(X)</code>.
    * </prerequisite>
    *  <code>reverse_iter([X<sub><math>1</math></sub>, ...,
    *    X<sub><math>n</math></sub>], F)</code> calls
    *  <code>F:do(X<sub><math>n</math></sub>)</code>, ...,
    *  <code>F:do(X<sub><math>1</math></sub>)</code>.
    */

   reverse_iter([], _).
   reverse_iter([H | T], F) :-
      reverse_iter(T, F),
      F:do(H).

   /** morph(List0, F, List1).
    * <prerequisite>
    * <code>List0</code> is a list.
    * <code>F</code> is a module which implements a predicate <code>do(X, Y)</code>.
    * </prerequisite>
    * <code>morph([X<sub><math>1</math></sub>, ...,
    *    X<sub><math>n</math></sub>], F, List1)</code> succeeds
    * if <code>List1</code> is the concatenation of
    * the lists <code>L<sub><math>1</math></sub>, ...,
    *    L<sub><math>n</math></sub></code> where
    * <code>F:do(X<sub><math>1</math></sub>,
    *   L<sub><math>1</math></sub>)</code>, ...,
    * <code>F:do(X<sub><math>n</math></sub>,
    *   L<sub><math>n</math></sub>)</code>.
    */

   morph([], _, []).
   morph([H | T], F, List) :-
      F:do(H, P),
      append(P, Tail, List),
      morph(T, F, Tail).

   /** fold_left(List, Initial, F, Final).
    * <prerequisite>
    * <code>List</code> is a list.
    * <code>F</code> is a module which implements a predicate <code>do(Item, Old, New)</code>.
    * </prerequisite>
    * <code>fold_left([X<sub><math>1</math></sub>, ...,
    *    X<sub><math>n</math></sub>],
    *    V<sub><math>0</math></sub>, F, V<sub><math>n</math></sub>)</code> succeeds
    * if <code>F:(X<sub><math>1</math></sub>, V<sub><math>0</math></sub>, V<sub><math>1</math></sub>), ...,
    * F:(X<sub><math>n</math></sub>, V<sub><math>n-1</math></sub>, V<sub><math>n</math></sub>)</code>.
    */

   fold_left([], Value, _) = Value.
   fold_left([H | T], Value, F) = 'data.list':fold_left(T, F:do(H, Value), F).

   /** fold_right(List, Initial, F, Final).
    * <prerequisite>
    * <code>List</code> is a list.
    * <code>F</code> is a module which implements a predicate <code>do(Item, Old, New)</code>.
    * </prerequisite>
    * <code>fold_left([X<sub><math>1</math></sub>, ...,
    *    X<sub><math>n</math></sub>],
    *    V<sub><math>n</math></sub>, F, V<sub><math>0</math></sub>)</code> succeeds
    * if <code>F:(X<sub><math>n</math></sub>, V<sub><math>n</math></sub>, V<sub><math>n-1</math></sub>), ...,
    * F:(X<sub><math>1</math></sub>, V<sub><math>1/math></sub>, V<sub><math>0</math></sub>)</code>.
    */

   fold_right([], Value, _) = Value.
   fold_right([H | T], Value, F) = F:do(H, 'data.list':fold_right(T, Value, F)).

   /** iter_fold(List0, List1, Initial, F, Final).
    * <prerequisite>
    * <code>List0</code> is a list.
    * <code>F</code> is a module which implements a predicate
    * <code>do(Item0, Item1, Old, New)</code>.
    * </prerequisite>
    * <code>iter_fold([X<sub><math>1</math></sub>, ...,
    *    X<sub><math>n</math></sub>],
    *    [Y<sub><math>1</math></sub>, ...,
    *    Y<sub><math>n</math></sub>],
    *    V<sub><math>0</math></sub>, F, V<sub><math>n</math></sub>)</code>
    * succeeds if
    * <code>F:(X<sub><math>1</math></sub>, Y<sub><math>1</math></sub>, V<sub><math>0</math></sub>, V<sub><math>1</math></sub>), ...,
    * F:(X<sub><math>n</math></sub>, Y<sub><math>n/math></sub>, V<sub><math>n-1</math></sub>, V<sub><math>n</math></sub>)</code>.
    */

   iter_fold([], [], Value, _) = Value.
   iter_fold([H0 | T0], [H1 | T1], Value, F) =
      'data.list':iter_fold(T0, T1, F:do(H0, H1, Value), F).

   zip([], []) = [].
   zip([X | Xs], [Y | Ys]) = [(X, Y) | zip(Xs, Ys)].

   /** accumulate(Generator, List).
    * <prerequisite>
    * <code>Generator</code> is a module which implements a predicate <code>get(X)</code>.
    * </prerequisite>
    * Calls <code>Generator:get(X<sub><math>i</math></sub>)</code> with a fresh variable
    * <code>X<sub><math>i</math></sub></code> at each call until failure.  The predicate
    * succeeds if <code>List</code> is the list
    * <code>[X<sub><math>1</math></sub>, ..., X<sub><math>n</math></sub>]</code>,
    * from the first introduced variable to the last.
    */

   accumulate(Generator, List) :-
      (
         Generator:get(H)
      ->
         List = [H | T],
         accumulate(Generator, T)
      ;
         List = []
      ).

   accumulate_sublists(Generator, List) :-
      (
         Generator:get(H)
      ->
         append(H, T, List),
         accumulate_sublists(Generator, T)
      ;
         List = []
      ).

   /** init(I, J, Init, List).
    * <prerequisite>
    * <code>I</code> and <code>J</code> are integers.
    * </prerequisite>
    * <prerequisite>
    * <code>Init</code> is a module which implements a predicate <code>get(N, X)</code>.
    * </prerequisite>
    * If <code>I</code> &le; <code>J</code>, calls
    * <code>Init:get(N, X<sub>N</sub>)</code> with a fresh variable
    * <code>X<sub>N</sub></code> for <code>N</code> from <code>I</code> to
    * <code>J</code>.  The predicate
    * succeeds if <code>List</code> is the list
    * <code>[X<sub><math>I</math></sub>, ..., X<sub><math>J</math></sub>]</code>.
    * <code>List</code> is empty if <code>I &gt; J</code>.
    */

   init(I, J, Init, List) :-
      (
         'data.number':(I =< J)
      ->
         List = [Init:get(I) | Tail],
         init('data.number':succ(I), J, Init, Tail)
      ;
         List = []
      ).

   /** split_at_delim_left(List, Delimiter, Left, Right).
    * Succeeds if <code>List</code> is the concatenation of
    * <code>Left</code>, <code>[Delimiter]</code>, <code>Right</code>
    * and <code>Delimiter</code> does not appear in <code>Left</code>.
    */

   split_at_delim_left([H | T], Delim, L0, L1) :-
      (
         H = Delim
      ->
         L0 = [],
         L1 = T
      ;
         L0 = [H | T0],
         split_at_delim_left(T, Delim, T0, L1)
      ).

   /** split(List, Delimiter, ListOfLists).
    * Succeeds if <code>List</code> can be splitted in sub-lists
    * separated by the element <code>Delimiter</code> and
    * <code>ListOfLists</code> enumerates these sub-lists;
    * <code>Delimiter</code> does not appear in sub-lists.
    */

   split(L, Delim, LL) :-
      (
         split_at_delim_left(L, Delim, H, T)
      ->
         LL = [H | TL],
         split(T, Delim, TL)
      ;
         LL = [L]
      ).

   /** merge(L0, L1, Preorder, L).
    * <prerequisite>
    * <code>Preorder</code> is a module with a predicate
    * <code>compare(X, Y, R)</code> which implements a total preorder over the elements of the lists <code>L0</code> and <code>L1</code>: <code>R</code>
    * is unified to <code>&lt;</code>, or <code>=</code>, or <code>&gt;</code>
    * whether <code>X</code> is less than, or equal to, or greather than <code>Y</code> respectively.
    * </prerequisite>
    * <prerequisite>
    * <code>L0</code> and <code>L1</code> is sorted increasingly with respect to this order.
    * </prerequisite>
    * <code>merge(L0, L1, Preorder, L)</code>
    * succeeds if <code>L</code> is the permutation of the concatenation
    * of <code>L0</code> and <code>L1</code> which is sorted with respect
    * <code>Preorder</code> and where the relative position of equivalent
    * elements with respect to <code>Preorder</code> is preserved,
    * elements of <code>L0</code> first. */

   merge(L0, L1, Preorder, L) :-
      (
         L0 = [H0 | T0]
      ->
         (
            L1 = [H1 | T1]
         ->
            Preorder:compare(H0, H1, R),
            (
               R = '>'
            ->
               L = [H1 | T],
               merge(L0, T1, Preorder, T)
            ;
               L = [H0 | T],
               merge(T0, L1, Preorder, T)
            )
         ;
            L1 = [],
            L = L0
         )
      ;
         L0 = [],
         L = L1
      ).

   /** merge_sort(List, Preorder, Sorted).
    * <prerequisite>
    * <code>List</code> is a list.
    * </prerequisite>
    * <prerequisite>
    * <code>Preorder</code> is a module with a predicate
    * <code>compare(X, Y, R)</code> which implements a total preorder over the elements of the lists <code>L0</code> and <code>L1</code>: <code>R</code>
    * is unified to <code>&lt;</code>, or <code>=</code>, or <code>&gt;</code>
    * whether <code>X</code> is less than, or equal to, or greather than <code>Y</code> respectively.
    * </prerequisite>
    * <code>merge_sort(List, Preorder, Sorted).</code> succeeds if <code>Sorted</code> is the permutation of <code>List</code>
    * stably sorted with respect to <code>Preorder</code>.
    * Runs in <math>O(n <log>n</log>)</math> where <math>n</math> is the length of <code>List</code>.*/

   merge_sort(List, Preorder, Sorted) :-
      (
         (
            List = []
         ;
            List = [_]
         )
      ->
         Sorted = List
      ;
         Singletons = 'data.list':iter(List) [] {
            do(X) = [X].
         },
         module Merge [Preorder] {
            do(L0, L1) :-
               (
                  (
                     L0 = []
                  ;
                     L0 = [_]
                  )
               ->
                  L1 = L0
               ;
                  L0 = [A, B | T0]
               ->
                  L1 = ['data.list':merge(A, B, Preorder) | T1],
                  do(T0, T1)
               ).
         },
         module Aux [Merge, Sorted] {
            do(LL) :-
               (
                  LL = [Sorted0]
               ->
                  Sorted = Sorted0
               ;
                  do(Merge:do(LL))
               ).
         },
         Aux:do(Singletons)
      ).

   /** split(List, ListTrue, ListFail, Select).
    * <prerequisite>
    * <code>List</code> is a list.
    * </prerequisite>
    * <prerequisite>
    * <code>Select</code> is a module with a predicate
    * <code>do(X)</code> succeeding or failing on elements of <code>List</code>.
    * </prerequisite> 
    * <code>ListTrue</code> is unified to the sub-list of <code>List</code> of elements
    * <code>X</code> such that <code>Select:do(X)</code> succeeds.
    * <code>ListFail</code> is unified to the sub-list of <code>List</code> of elements
    * <code>Y</code> such that <code>Select:do(Y)</code> fails.
    */

   split([], [], [], _).
   split([H | T], ListTrue, ListFail, Select) :-
      (
         Select:do(H)
      ->
         ListTrue = [H | TT],
         split(T, TT, ListFail, Select)
      ;
         ListFail = [H | TF],
         split(T, ListTrue, TF, Select)
      ).

   /** quick_sort(List, Preorder, Sorted).
    * <prerequisite>
    * <code>List</code> is a list.
    * </prerequisite>
    * <prerequisite>
    * <code>Preorder</code> is a module with a predicate
    * <code>compare(X, Y, R)</code> which implements a total preorder over the elements of the lists <code>L0</code> and <code>L1</code>: <code>R</code>
    * is unified to <code>&lt;</code>, or <code>=</code>, or <code>&gt;</code>
    * whether <code>X</code> is less than, or equal to, or greather than <code>Y</code> respectively.
    * </prerequisite>
    * <code>quick_sort(List, Preorder, Sorted).</code> succeeds if <code>Sorted</code> is the permutation of <code>List</code>
    * stably sorted with respect to <code>Preorder</code>.
    * Runs in <math>O(n <sup>2</sup>)</math> where <math>n</math> is the length of <code>List</code>.*/

   quick_sort(List, Preorder, Sorted) :-
      (
         (
            List = []
         ;
            List = [_]
         )
      ->
         Sorted = List
      ;
         List = [P | T],
         split(T, List0, List1) [Preorder, P] {
            do(X) :-
               Preorder:compare(X, P, '<').
         },
         quick_sort(List0, Preorder, Sorted0),
         quick_sort(List1, Preorder, Sorted1),
         append(Sorted0, [P | Sorted1], Sorted)
      ).

   /** longuest_prefix(List, Prefix).
    *  Succeeds if <code>Prefix</code> is a prefix of <code>List</code>,
    *  preferably the longuest one.
    */

   longuest_prefix([H | T0], [H | T1]) :-
      longuest_prefix(T0, T1).
   longuest_prefix([], _).

   /** subst(Src, Table, Tgt).
       <prerequisite>
       <code>Src</code> is a list.
       </prerequisite>
       <prerequisite>
       <code>Table</code> is a list with elements of the form
       <code>A => B</code> where <code>A</code> and <code>B</code>
       are lists.
       </prerequisite>
       Succeeds if <code>Tgt</code> is the list obtained from
       <code>Src</code> by substituting every occurrences of
       sub-list <code>A</code> in <code>Src</code> by <code>B</code>,
       for every <code>A => B</code> in <code>Table</code>.
    */

   subst(Src, Table, Tgt) :-
      (
         Src = []
      ->
         Tgt = []
      ;
         module [Table] {
            find([], [H | SrcTail], [H | TgtTail]) :-
               subst(SrcTail, Table, TgtTail).
            find([From => To | TableTail], Src, Tgt) :-
               (
                  append(From, SrcTail, Src)
               ->
                  append(To, TgtTail, Tgt),
                  subst(SrcTail, Table, TgtTail)
               ;
                  find(TableTail, Src, Tgt)
               ).
         }:find(Table, Src, Tgt)
      ).

   delete([], _, []).
   delete([H | T], X, L) :-
      (
         X = H
      ->
         L = U
      ;
         L = [H | U]
      ),
      delete(T, X, U).

   select(X, [X | T], T).
   select(X, [Y | T], [Y | U]) :-
      select(X, T, U).

   member_partial(X, L, New) :-
      (
         data.var:check(L)
      ->
         New = 'true',
         L = [X | _]
      ;
         L = [X | _]
      ->
         true
      ;
         L = [_ | T],
         member_partial(X, T, New)
      ).

   nth1(N, L, X) :-
      kernel:list_nth1(N, L, X).

   index1(X, L, N) :-
      nth1(N, L, X).

   stream(List) =
      (Contents = data.ref.backtrackable:new(List)) >>
      module {
         read(H) :-
            (
               Contents:get = [H | T]
            ->
               Contents:set(T)
            ;
               H = 'end_of_file'
            ).
      }.

   random_item(Generator, List) =
      nth1(Generator:random(1, length(List)), List).
}
