module emop.sgn [SignatureCacheRef, ConsultSignatureRef, SignatureResultRef] {
   constructor: special/4, syntax/3, level/3, notation/3, link/3, clause/6, symbol/4,
      symbol/5,
      file/2, error/2, none/0, operator/0, number/3, empty/2, doc/4, doc/3,
      unescaped/0, var/1, variable/0, 'module'/7.

   from_mod_spec(ModSpec, Pos, Sgn) :-
      (
         ModSpec = 'name'(ModName, _, _),
         data.list:member((ModName, Sgn), ConsultSignatureRef:get)
      ->
         true
      ;
         emop.compile:lookup(
            ModSpec,
            Pos,
            BaseFilename,
            Filename,
            Stream
         ),
         from_file(BaseFilename, Filename, Stream, Sgn),
         Stream:close
      ).

   add_consult_sgn(ModName, Sgn) :-
      ConsultSignatureRef:set([(ModName, Sgn) | ConsultSignatureRef:get]).

   cache(Filename, ChangedDate, Sgn) :-
      SignatureCacheRef:set(['cache'(Filename, ChangedDate, Sgn) | SignatureCacheRef:get]).

   from_file(BaseFilename, Filename, Stream, Sgn) :-
      ChangedDate = sys.file:last_modification(Filename),
      (
         data.list:member('cache'(Filename, ParsedDate, Sgn0), SignatureCacheRef:get),
         data.term:(ChangedDate =< ParsedDate)
      ->
         Sgn = Sgn0
      ;
         (
            data.var:check(SignatureResultRef)
         ->
            SignatureResultRef = data.ref.non_backtrackable:new([])
         ;
            true
         ),
         (
            (
               emop.compile:read_interface(BaseFilename, Filename, _, _, _, Sgn0, _, _, _)
            ->
               true
            ;
               Sgn0 = from_stream(Filename, Stream)
            ),
            SignatureResultRef:set(Sgn0),
            fail
         ;
            Sgn0 = SignatureResultRef:get,
            SignatureResultRef:set([]),
            cache(Filename, ChangedDate, Sgn0),
            Sgn = Sgn0
         )
      ).

   from_stream(Filename, Stream, Sgn) :-
      emop.parser:parse_file(Filename, Stream, this)
      | from_goal('none', Sgn, _).

   constructor: callable/5, '=>'/2, atom/3.

   from_goal(callable(F, Escaped, Args, _Pb, _Pe), Ctx, ModuleSgn, Sgn) :-
      !,
      from_goal_callable(F, Escaped, Args, Ctx, ModuleSgn, Sgn).
   from_goal(module(Name, Env, _, Sgn0, _, _, _), Ctx, ModuleSgn, Sgn) :-
      !,
      from_module(Name, Env, Sgn0, _, Ctx, ModuleSgn, Sgn).
   from_goal(_, _Ctx, _ModuleSgn, _Sgn).

   from_goal_callable(',', 'operator', [A, B], Ctx, ModuleSgn, Sgn) :-
      !,
      from_goal(A, Ctx, ModuleSgn, Sgn),
      from_goal(B, Ctx, ModuleSgn, Sgn).
   from_goal_callable(_, _, _, _, _, _).

   from_module('atom'(Name, Escaped, _, _), _Env, Sgn0, _, _Ctx, ModuleSgn, _Sgn) :-
      !,
      (
         emop.parser:is_variable(Name, Escaped, 'none')
      ->
         true
      ;
         ModuleSgn = Sgn0
      ).
   from_module(_, _Env, _Body, _Ctx, _ModuleSgn, _Sgn).

   special_head(
      Symbol,
      [
         callable(Symbol, unescaped, none, _, _),
         callable(':', unescaped, none, _, _)
         |
         Tail
      ],
      Tail
   ).

   directive_syntax(Ast, Directive) :-
      Concat = emop.parse:reflat(Ast),
      List = emop.parse:enum_concat_list(Concat),
      special_head(Symbol, List, Tail),
      Symbol match (
         (
            'syntax'
         ;
            'constructor'
         ;
            'abstract'
         ;
            'import'
         ;
            'notation'
         ;
            'level'
         ;
            'link'
         )
      ->
         !,
         file(Pb, Pe) = emop.parse:position(Ast),
         Directive = special(Symbol, Tail, Pb, Pe)
      ).

   directive_sgn(special(Symbol, S, Pb, Pe), Sgn) =
      module {
        do('syntax') = [syntax(S, Pb, Pe) | Sgn].

        do('level') = [level(S, Pb, Pe) | Sgn].

        do('constructor') =
           data.list:fold_left(read_sgn(S, Pb, Pe), Sgn) [] {
              do(symbol(F, N, Pb, Pe), Sgn, NewSgn) :-
                 M = data.number:(N + 1),
                 (
                    data.list:member(symbol(F, M, _Kind, _, _), Sgn)
                 ->
                    throw(
                       error(
                          'This symbol is already defined',
                          file(Pb, Pe)
                       )
                    )
                 ;
                    NewSgn = [symbol(F, M, 'constructor', none, file(Pb, Pe)) | Sgn]
                 )   
           }.

        do('abstract') =
           data.list:fold_left(read_sgn(S, Pb, Pe), Sgn) [] {
              do(symbol(F, N, Pb, Pe), Sgn, NewSgn) :-
                 (
                    data.list:member(symbol(F, N, _Kind, _, _), Sgn)
                 ->
                    throw(
                       error(
                          'This symbol is already defined',
                          file(Pb, Pe)
                       )
                    )
                 ;
                    NewSgn = [symbol(F, N, 'abstract', none, file(Pb, Pe)) | Sgn]
                 )   
           }.

        do('import') = Sgn.

        do('notation') = [notation(S, Pb, Pe) | Sgn].

        do('link') = [link(S, Pb, Pe) | Sgn].
     }:do(Symbol).

   enum_clauses(Ast, Clauses) :-
      NewClausesRef = 'data.ref.backtrackable':new([]),
      module EnumClauses [
         EnumClauses,
         NewClausesRef,
      ] {
         expand_clauses(callable(F, Escaped, Args, Pb, Pe)) :-
            expand_clauses_callable(F, Escaped, Args, Pb, Pe, 'nodoc').
         expand_clauses('()'(Sub_clauses, _, _)) :-
            expand_clauses(Sub_clauses).
         expand_clauses(module(_, _, _, _, _, Pb, Pe)) :-
            throw(error('A module is not a valid clause', file(Pb, Pe))).
         expand_clauses(number(_, Pb, Pe)) :-
            throw(error('A number is not a valid clause', file(Pb, Pe))).
         expand_clauses(empty(_, _)).
         expand_clauses(doc(Comment, Pb, Pe, Clause)) :-
            expand_clause_doc(Clause, doc(Comment, Pb, Pe)).
         
         expand_clause_doc(callable(F, Escaped, Args, Pb, Pe), Doc) :-
            expand_clauses_callable(F, Escaped, Args, Pb, Pe, Doc).
         expand_clause_doc('()'(Sub_clauses, _, _), Doc) :-
            expand_clause_doc(Sub_clauses, Doc).
         expand_clause_doc(module(_, _, _, _, _, Pb, Pe), _) :-
            throw(error('A module is not a valid clause', file(Pb, Pe))).
         expand_clause_doc(number(_, Pb, Pe), _) :-
            throw(error('A number is not a valid clause', file(Pb, Pe))).
         expand_clause_doc(empty(Pb, Pe), _) :-
            throw(error('A documentation should be followed by an implementation', file(Pb, Pe))).
         expand_clause_doc(doc(_, Pb, Pe, _), _) :-
            throw(error('Predicate already documented', file(Pb, Pe))).

         expand_xml(Atom, Xml) :-
            data.atom.(subst(
               Atom,
               ['<' => '&lt;', '>' => '&gt;', '\"' => '&quot;', '\'' => '&apos;'],
               Xml
            )).

         reserved(',', 2).
         reserved('->', 2).
         reserved(';', 2).
         reserved(':', 2).
         reserved('.', 2).
         reserved('=', 2).
         reserved('', 2).
         reserved(',', 3).
         reserved('->', 3).
         reserved(';', 3).
         reserved(':', 3).
         reserved('.', 3).
         reserved('=', 3).
         reserved('.', 0).
         reserved('catch', 3).
         reserved('call_det', 2).
         reserved('true', 0).
         reserved('fail', 0).
         reserved('repeat', 0).
         reserved('for', 3).
         reserved('throw', 1).

         add_clause(F, Args, B, Doc, Pb, Pe) :-
            data.list:length(Args, N),
            (
               reserved(F, N)
            ->
               throw(error(data.atom:format('This clause redefines the reserved predicate ~a/~d.', [F, N]), file(Pb, Pe)))
            ;
               NewClausesRef:set(
                  [clause(F, Args, B, Doc, Pb, Pe) | NewClausesRef:get]
               )
            ).

         check_predicate_quotes(Escaped, Pb, Pe) :-
            (
               Escaped = 'unescaped'
            ->
               true
            ;
               emop.message:warning('Predicates should be quoted between `...\'', file(Pb, Pe))
            ).

         expand_clauses_callable(':-', operator, [H, B], Pb, Pe, Doc) :-
            !,
            decompose_head(H, _Pb, _Pe, F, Escaped, Args),
            (
               F = '=',
               Escaped = operator,
               Args = [A, R]
            ->
               emop.message:warning('Goals in function declaration are deprecated', file(Pb, Pe)),
               EnumClauses:expand_eq_clause_deprecated(A, R, Pb, Pe, B, Doc)
            ;
               check_predicate_quotes(Escaped, Pb, Pe),
               (
                  Args = none
               ->
                  Args0 = []
               ;
                  Args0 = Args
               ),
               add_clause(F, Args0, B, Doc, Pb, Pe)
            ).
         expand_clauses_callable(F = ('$~$'; '='; ','; ':'), operator, [A, B], Pb, Pe, _Doc) :-
            Special = directive_syntax(callable(F, operator, [A, B], Pb, Pe)),
            !,
            NewClausesRef:set(
               [
                  Special
                  |
                  NewClausesRef:get
               ]
            ).
         expand_clauses_callable('=', operator, [A, R], Pb, Pe, Doc) :-
            !,
            EnumClauses:expand_eq_clause(A, R, Pb, Pe, Doc).
         expand_clauses_callable('.', operator, [A0, A1], Pb, Pe, Doc) :-
            !,
            (
               Doc = 'nodoc'
            ->
               expand_clauses(A0),
               expand_clauses(A1)
            ;
               throw(
                  error(
                     'A documentation should be followed by an implementation',
                     file(Pb, Pe)
                  )
               )
            ).
         expand_clauses_callable(F, Escaped, Args, Pb, Pe, Doc) :-
            check_predicate_quotes(Escaped, Pb, Pe),
            (
               Args = none
            ->
               Args0 = []
            ;
               Args0 = Args
            ),
            add_clause(F, Args0, callable('true', unescaped, none, Pb, Pe), Doc, Pb, Pe).
      
         expand_eq_clause(A, R, Pb, Pe, Doc) :-
            module Aux [Aux, F, Args] {
               do(callable(F, Escaped, Args, Pb, Pe)) :-
                  check_predicate_quotes(Escaped, Pb, Pe).
               do('()'(Sub_clauses, _, _)) :-
                  Aux:do(Sub_clauses).
               do(callable(':', operator, [_, _], Pb, Pe)) :-
                  throw(
                     error(
                        'Predicate definition cannot be prefixed',
                        file(Pb, Pe)
                     )
                  ).
               do(module(_, _, _, _, _, Pb, Pe)) :-
                  throw(
                     error(
                        'A module is not a valid predicate',
                        file(Pb, Pe)
                     )
                  ).
               do(number(_, Pb, Pe)) :-
                  throw(
                     error(
                        'A number is not a valid predicate',
                        file(Pb, Pe)
                     )
                  ).
               do(empty(Pb, Pe)) :-
                  throw(
                     error(
                        'An empty item is not a valid predicate',
                        file(Pb, Pe))
                  ).
               do(doc(_, Pb, Pe, _)) :-
                  throw(
                     error('Documentation not allowed here', file(Pb, Pe))
                  ).
            },
            Aux:do(A),
            (
               Args = none
            ->
               Args0 = [R]
            ;
               data.list:append(Args, [R], Args0)
            ),
            add_clause(F, Args0, callable('true', unescaped, [], Pb, Pe), Doc, Pb, Pe).
      
         expand_eq_clause_deprecated(A, R, Pb, Pe, B, Doc) :-
            module Aux [Aux, F, Args] {
               do(callable(F, _, Args, _, _)).
               do('()'(Sub_clauses, _, _)) :-
                  Aux:do(Sub_clauses).
               do(callable(':', operator, [_, _], Pb, Pe)) :-
                  throw(
                     error(
                        'Predicate definition cannot be prefixed',
                        file(Pb, Pe)
                     )
                  ).
               do(module(_, _, _, _, _, Pb, Pe)) :-
                  throw(
                     error(
                        'A module is not a valid predicate',
                        file(Pb, Pe)
                     )
                  ).
               do(number(_, Pb, Pe)) :-
                  throw(
                     error(
                        'A number is not a valid predicate',
                        file(Pb, Pe)
                     )
                  ).
               do(empty(Pb, Pe)) :-
                  throw(
                     error(
                        'An empty item is not a valid predicate',
                        file(Pb, Pe))
                  ).
               do(doc(_, Pb, Pe, _)) :-
                  throw(
                     error('Documentation not allowed here', file(Pb, Pe))
                  ).
            },
            Aux:do(A),
            emop.parser:fresh_var(V),
            V = var(FreshVarName),
            V0 = callable(FreshVarName, variable, none, Pb, Pe),
            B0 = callable(',', operator, [B, callable('=', operator, [V0, R], Pb, Pe)], Pb, Pe),
            (
               Args = none
            ->
               Args0 = [V0]
            ;
               data.list:append(Args, [V0], Args0)
            ),
            add_clause(F, Args0, B0, Doc, Pb, Pe).
      
         decompose_head(
            callable(F, Escaped, Args, Pb, Pe), Pb, Pe, F, Escaped, Args0
         ) :-
            (
               Args = none
            ->
               Args0 = []
            ;
               Args0 = Args
            ).
         decompose_head('()'(SubHead, _, _), Pb, Pe, F, Escaped, Args) :-
            EnumClauses:decompose_head(SubHead, Pb, Pe, F, Escaped, Args).
         decompose_head(module(_, _, _, _, _, Pb, Pe), _, _, _, _, _) :-
            throw(error('A module is not a valid head', file(Pb, Pe))).
         decompose_head(number(_, Pb, Pe), _, _, _, _, _) :-
            throw(error('A number is not a valid head', file(Pb, Pe))).
         decompose_head(empty(Pb, Pe), _, _, _, _, _) :-
            throw(error('Empty head not allowed', file(Pb, Pe))).
         decompose_head(doc(_, Pb, Pe, _), _, _, _, _, _) :-
            throw(
               error('Documentation not allowed here', file(Pb, Pe))
            ).
      }:expand_clauses(Ast),
      Clauses = 'data.list':reverse(NewClausesRef:get).

  read_sgn(Constr, Pb, Pe) =
     module [Pb, Pe] {
        atomic(F, callable(F, _, none, _, _)) :- !.
        atomic(F, '()'(atomic(F), _, _)).

        do([], Sgn, Sgn) :- !.
        do([atomic(F), callable('/', 'unescaped', none, Pb0, Pe0), number(N, _, _) | Tail], Sgn, NewSgn) :-
           !,
           (
              Tail = [callable(',', 'unescaped', none, _, _) | Tail0]
           ->
              true
           ;
              Tail0 = Tail
           ),
           do(Tail0, [symbol(F, N, Pb0, Pe0) | Sgn], NewSgn).
        do(S, _, _) :-
           io.err:print_endline(S),
           throw(
              error(
                 'Invalid signature',
                 file(Pb, Pe)
              )
           ).
     }:do(Constr, []).

  extract_sgn(Clauses, Sgn) :-
     (
     Sgn = data.list:reverse(data.list:fold_left(Clauses, []) [] {
        do(special(S, Desc, Pb, Pe), Sgn, NewSgn) :-
           NewSgn = directive_sgn(special(S, Desc, Pb, Pe), Sgn).

        infer_args_info(Args, Body, Args0) :-
           module M [] {
              collect_goal(callable(F, E, A, _, _), List) :-
                 !,
                 collect_goal_callable(F, E, A, List).
              collect_goal(module(NameAtom, _, _, Sgn, _, _, _), List) :-
                 !,
                 NameAtom = 'atom'(A, E, _, _),
                 E \= 'escaped',
                 List = ['module'(A, Sgn)].
              collect_goal('()'(B, _, _)) :-
                 !,
                 collect_goal(B).
              collect_goal(_).

              collect_goal_callable(',', operator, [Left, Right], List) :-
                 !,
                 collect_goal(Left, List0),
                 collect_goal(Right, List1),
                 List = data.list.(List0 ++ List1).
              collect_goal_callable(':', operator, [Left, Right], List) :-
                 !,
                 Left = callable(LF, LEscaped, LArgs, _, _),
                 (
                    emop.parser:is_variable(LF, LEscaped, LArgs)
                 ->
                    Mod = var(LF)
                 ;
                    LArgs = none
                 ->
                    Mod = 'atom'(LF)
                 ),
                 Right = callable(RF, _REscaped, RArgs, _, _),
                 simplify_args(RArgs, RArgs', List0),
                 List = data.list.(List0 ++ ['call'(Mod, RF, RArgs')]).
              collect_goal_callable(F, unescaped, Args, List) :-
                 simplify_args(Args, Args', List0),
                 List = data.list.(List0 ++ ['call'('scope', F, Args')]).

              simplify_args([], [], []).
              simplify_args([H | T], [H' | T'], List) :-
                 (
                    H = callable(F, Escaped, Args, _, _),
                    emop.parser:is_variable(F, Escaped, Args)
                 ->
                    H' = 'var'(F),
                    List0 = []
                 ;
                    H = module(_, _, _, Sgn, _, _, _)
                 ->
                    H = emop.parser:fresh_var,
                    H = 'var'(X),
                    List0 = ['module'(X, Sgn)]
                 ;
                    H' = none,
                    List0 = []
                 ),
                 data.list:append(List0, List1, List),
                 simplify_args(T, T', List1).
           },
           M:simplify_args(Args, Args', List0),
           M:collect_goal(Body, List1),
           !,
           Args0 = 'suspended'(Args', data.list.(List0 ++ List1)).
        infer_args_info(Args, Body, Args0) :-
           Env = data.record:{},
           module {
              collect_goal(callable(F, E, A, _, _)) :-
                 !,
                 collect_goal_callable(F, E, A).
              collect_goal(module(NameAtom, _, _, Sgn, _, _, _)) :-
                 !,
                 collect_module(NameAtom, Sgn, _).
              collect_goal('()'(B, _, _)) :-
                 !,
                 collect_goal(B).
              collect_goal(_).

              unify(A, B) :-
                 (
                    (
                       data.var:check(A)
                    ;
                       data.var:check(B)
                    )
                 ->
                    A = B
                 ;
                    AR = emop.unionfind:get_root(A),
                    BR = emop.unionfind:get_root(B),
                    'root'(AS) = AR:get,
                    'root'(BS) = BR:get,
                    AR:set('root'(data.list:append(AS, BS))),
                    BR:set('link'(AR))
                 ).

              collect_module(NameAtom, Sgn, V) :-
                 V = data.ref.backtrackable:new('root'(Sgn)),
                 (
                    NameAtom = 'atom'(A, E, _, _),
                    E \= 'escaped'
                 ->
                    unify(Env:{A}, V)
                 ;
                    true
                 ).

              collect_goal_callable('=', operator, [callable(A, AE, none, _, _), callable(B, BE, none, _, _)]) :-
                 AE \= 'escaped', BE \= 'escaped',
                 !,
                 unify(Env:{A}, Env:{B}).
              collect_goal_callable('=', operator, [Left, Right]) :-
                 (
                    Left = callable(A, AE, none, _, _),
                    Right = module(NameAtom, _, _, Sgn, _, _, _)
                 ;
                    Left = module(NameAtom, _, _, Sgn, _, _, _),
                    Right = callable(A, AE, none, _, _)
                 ),
                 !,
                 unify(Env:{A}, collect_module(NameAtom, Sgn)).
              collect_goal_callable(',', operator, [Left, Right]) :-
                 !,
                 collect_goal(Left),
                 collect_goal(Right).
              collect_goal_callable(_, _, A) :-
                 collect_term_list(A).

              collect_term(callable(_, _, A, _, _)) :-
                 !,
                 collect_term_list(A).
              collect_term(module(NameAtom, _, _, Sgn, _, _, _)) :-
                 !,
                 collect_module(NameAtom, Sgn, _).
              collect_term('()'(B, _, _)) :-
                 !,
                 collect_goal(B).
              collect_term(_).

              collect_term_list(A) :-
                 (
                    A = none
                 ->
                    true
                 ;
                    data.list:iter(A) {
                       do(X) :- collect_term(X).
                    }
                 ).              
           }:collect_goal(Body),
           Args0 = data.list:iter(Args) {
              do(Arg) =
                 (
                    Arg = callable(A, AE, none, _, _),
                    AE \= 'escaped',
                    E = Env:{A},
                    \+ data.var:check(E),
                    'root'(S) = emop.unionfind:get_root(E):get
                 ->
                    S
                 ;
                    Arg = module(_, _, _, S, _, _, _)
                 ->
                    S
                 ;
                    none
                 ).
           }.

        do(clause(F, Args, B, _, Pb0, Pe0), Sgn, NewSgn) :-
           (
           data.list:length(Args, N),
           (
              'data.list':member(symbol(F, N, Kind, ArgsInfo, _), Sgn)
           ->
              (
                 Kind = 'predicate'
              ->
                 (
                    ArgsInfo = 'suspended'(_, _)
                 ->
                    true
                 ;
                    ArgsInfo0 = infer_args_info(Args, B),
                    (
                       ArgsInfo0 = 'suspended'(_, _)
                    ->
                       true
                    ;
                       module [] {
                          do([], []).
                          do([H0 | T0], ArgsInfo) :-
                             ArgsInfo = [H | T],
                             (
                                H0 = none
                             ->
                                true
                             ;
                                (
                                   H = none
                                ->
                                   NewH = H0
                                ;
                                   NewH = data.list:append(H0, H)
                                ),
                                data.term:setarg1(1, ArgsInfo, NewH),
                                do(T0, T)
                             ),
                             do(T0, T).
                       }:do(ArgsInfo0, ArgsInfo)
                    )
                 ),
                 NewSgn = Sgn
              ;
                 throw(
                    error(
                       'This symbol should not have any clause',
                       file(Pb0, Pe0)
                    )
                 )
              )
           ;
              ArgsInfo = infer_args_info(Args, B),
              NewSgn = [symbol(F, N, 'predicate', ArgsInfo, file(Pb0, Pe0)) | Sgn]
           )
           ->
              true
           ;
              throw(error('clause bug', file(Pb0, Pe0)))
           ).
     })
     ->
        true
     ;
        throw('extract_sgn fail')
     ).

},
SignatureCacheRef = data.ref.non_backtrackable:new([]),
ConsultSignatureRef = data.ref.backtrackable:new([])